/* Copyright (c) 2007-2014. The SimGrid Team.
 * All rights reserved.                                                     */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#ifndef MC_PRIVATE_H
#define MC_PRIVATE_H

#include "simgrid_config.h"
#include <stdio.h>
#include <stdbool.h>
#ifndef WIN32
#include <sys/mman.h>
#endif
#include <elfutils/libdw.h>

#include "mc/mc.h"
#include "mc_base.h"
#include "mc/datatypes.h"
#include "xbt/fifo.h"
#include "xbt/config.h"
#include "xbt/function_types.h"
#include "xbt/mmalloc.h"
#include "../simix/smx_private.h"
#include "../xbt/mmalloc/mmprivate.h"
#include "xbt/automaton.h"
#include "xbt/hash.h"
#include "msg/msg.h"
#include "msg/datatypes.h"
#include "xbt/strbuff.h"
#include "xbt/parmap.h"

#include "mc_forward.h"
#include "mc_mmu.h"
#include "mc_page_store.h"
#include "mc_interface.h"
#include "mc_mmalloc.h"
#include "mc_model_checker.h"
#include "mc_snapshot.h"
#include "mc_object_info.h"
#include "mc_location.h"
#include "mc_state.h"
#include "mc_comm_pattern.h"
#include "mc_request.h"
#include "mc_safety.h"
#include "mc_liveness.h"
#include "mc_memory_map.h"

SG_BEGIN_DECL()

typedef struct s_mc_function_index_item s_mc_function_index_item_t, *mc_function_index_item_t;

/****************************** Snapshots ***********************************/

extern xbt_dynar_t mc_checkpoint_ignore;

/********************************* MC Global **********************************/

extern FILE *dot_output;
extern const char* colors[13];
extern xbt_parmap_t parmap;

extern int user_max_depth_reached;

int MC_deadlock_check(void);
void MC_replay(xbt_fifo_t stack, int start);
void MC_replay_liveness(xbt_fifo_t stack, int all_stack);
void MC_show_deadlock(smx_simcall_t req);
void MC_show_stack_safety(xbt_fifo_t stack);
void MC_dump_stack_safety(xbt_fifo_t stack);

/** Stack (of `mc_state_t`) representing the current position of the
 *  the MC in the exploration graph
 *
 *  It is managed by its head (`xbt_fifo_shift` and `xbt_fifo_unshift`).
 */
extern xbt_fifo_t mc_stack;

int get_search_interval(xbt_dynar_t list, void *ref, int *min, int *max);


/****************************** Statistics ************************************/

typedef struct mc_stats {
  unsigned long state_size;
  unsigned long visited_states;
  unsigned long visited_pairs;
  unsigned long expanded_states;
  unsigned long expanded_pairs;
  unsigned long executed_transitions;
} s_mc_stats_t, *mc_stats_t;

extern mc_stats_t mc_stats;

void MC_print_statistics(mc_stats_t stats);

extern char *libsimgrid_path;

/********************************** Snapshot comparison **********************************/

typedef struct s_mc_comparison_times{
  double nb_processes_comparison_time;
  double bytes_used_comparison_time;
  double stacks_sizes_comparison_time;
  double binary_global_variables_comparison_time;
  double libsimgrid_global_variables_comparison_time;
  double heap_comparison_time;
  double stacks_comparison_time;
}s_mc_comparison_times_t, *mc_comparison_times_t;

extern __thread mc_comparison_times_t mc_comp_times;
extern __thread double mc_snapshot_comparison_time;

int snapshot_compare(void *state1, void *state2);
void print_comparison_times(void);

//#define MC_DEBUG 1
#define MC_VERBOSE 1

/********************************** Variables with DWARF **********************************/

dw_frame_t MC_find_function_by_ip(void* ip);
mc_object_info_t MC_ip_find_object_info(void* ip);

void MC_find_object_address(memory_map_t maps, mc_object_info_t result);

/********************************** Miscellaneous **********************************/

typedef struct s_local_variable{
  dw_frame_t subprogram;
  unsigned long ip;
  char *name;
  dw_type_t type;
  void *address;
  int region;
}s_local_variable_t, *local_variable_t;

/* *********** Sets *********** */

typedef struct s_mc_address_set *mc_address_set_t;

mc_address_set_t mc_address_set_new();
void mc_address_set_free(mc_address_set_t* p);
void mc_address_add(mc_address_set_t p, const void* value);
bool mc_address_test(mc_address_set_t p, const void* value);

/* *********** Hash *********** */

/** \brief Hash the current state
 *  \param num_state number of states
 *  \param stacks stacks (mc_snapshot_stak_t) used fot the stack unwinding informations
 *  \result resulting hash
 * */
uint64_t mc_hash_processes_state(int num_state, xbt_dynar_t stacks);

/* *********** Snapshot *********** */

#define MC_LOG_REQUEST(log, req, value) \
  if (XBT_LOG_ISENABLED(log, xbt_log_priority_debug)) { \
    char* req_str = MC_request_to_string(req, value); \
    XBT_DEBUG("Execute: %s", req_str); \
    xbt_free(req_str); \
  }

/** @brief Dump the stacks of the application processes
 *
 *   This functions is currently not used but it is quite convenient
 *   to call from the debugger.
 *
 *   Does not work when an application thread is running.
 */
void MC_dump_stacks(FILE* file);

SG_END_DECL()

#endif
